package com.example.ueberwachungssystem.Detection;

import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.ImageFormat;
import android.media.Image;
import android.os.CountDownTimer;
import android.util.Log;
import android.widget.ImageView;

import androidx.annotation.NonNull;
import androidx.camera.core.CameraSelector;
import androidx.camera.core.ExperimentalGetImage;
import androidx.camera.core.ImageAnalysis;
import androidx.camera.core.Preview;
import androidx.camera.lifecycle.ProcessCameraProvider;
import androidx.core.content.ContextCompat;
import androidx.lifecycle.LifecycleOwner;

import com.google.common.util.concurrent.ListenableFuture;

import org.opencv.android.OpenCVLoader;
import org.opencv.android.Utils;
import org.opencv.core.Core;
import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.MatOfPoint;
import org.opencv.core.Scalar;
import org.opencv.imgproc.Imgproc;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.ExecutionException;


/**
 * Video Detector inherits some methods from abstract Detector class (more info there)
 *
 *
 * USE FROM MAIN ACTIVITY:
 *
 * VideoDetector vd = new VideoDetector(this);
 *
 * */


@ExperimentalGetImage
public class VideoDetector extends Detector {

    // Calling Activity
    private final Context context;
    // Camera Provider
    private ProcessCameraProvider cameraProvider;
    private Boolean isDetectionRunning = false;
    // Detection
    private Mat previousImage = null;
    private Mat lastThresh;
    private Mat currentOut;

    public ImageView imageView1 = null;
    public ImageView imageView2 = null;


    private int frameCnt = 0;



    // Parameters
    private static final float PIXEL_THRESHOLD = 30f;  // Luminosity (brightness channel of YUV_420_888)
    private static final int BLUR_KERNEL_SIZE = 5;
    private static final int DILATE_KERNEL_SIZE = 5;
    private static final float CONTOUR_THRESHOLD = 100;
    private static final float ALARM_THRESHOLD = 0.2f;   // Percent of pixels changed
    private static final long START_DELAY = 1;   // milliseconds
    private static final android.util.Size IMAGE_RES = new android.util.Size(640, 480);




    /**
     * Constructor
     * @param context: the context of calling activity (usually "this")
     * */
    public VideoDetector(Context context) {
        super();
        this.context = context;
    }


    /**
     *  Starts the Video Detection
     * */
    @Override
    public void startDetection() {
        if (isDetectionRunning)
            return;

        // Open CV startup check
        if (!OpenCVLoader.initDebug()) {
            Log.e("OpenCV", "Unable to load OpenCV!");
            return;
        } else
            Log.d("OpenCV", "OpenCV loaded Successfully!");


        // Request Camera Provider
        final ListenableFuture<ProcessCameraProvider> cameraProviderFuture = ProcessCameraProvider.getInstance(context);
        //Check for Camera availability
        cameraProviderFuture.addListener(() -> {
            try {
                cameraProvider = cameraProviderFuture.get();
                bindAnalysis(cameraProvider);
                isDetectionRunning = true;
            } catch (ExecutionException | InterruptedException e) {
                // No errors need to be handled for this Future. This should never be reached.
            }
        },ContextCompat.getMainExecutor(context));
    }

    /**
     * Stops the Video Detection
     * */
    @Override
    public void stopDetection() {
        if (!isDetectionRunning)
            return;
        cameraProvider.unbindAll();
        isDetectionRunning = false;
    }


    /**
     * Binds the Luminosity Analyzer (configure and run Analysis)
     * @param cameraProvider: CameraProvider of Context passed by Constructor
     * */
    private void bindAnalysis(@NonNull ProcessCameraProvider cameraProvider) {
        // Configure and create Image Analysis
        ImageAnalysis.Builder builder = new ImageAnalysis.Builder();
        builder.setTargetResolution(IMAGE_RES);
        builder.setBackpressureStrategy(ImageAnalysis.STRATEGY_KEEP_ONLY_LATEST);
        builder.setOutputImageFormat(ImageAnalysis.OUTPUT_IMAGE_FORMAT_YUV_420_888);
        ImageAnalysis imageAnalysis = builder.build();

        // Set Analyzer
        imageAnalysis.setAnalyzer(ContextCompat.getMainExecutor(context), imageProxy -> {
            if (imageProxy.getFormat() == ImageFormat.YUV_420_888) {
                Image image = imageProxy.getImage();
                assert image != null;

                Mat mat = extractYChannel(image);

                debugMat(mat, imageView2);

                //mat = alternativeProcessImage(mat);
                mat = processImage(mat);
                debugMat(mat, imageView1);
            }
            imageProxy.close();
        });
        // Create Preview
        Preview preview = new Preview.Builder().build();
        // Specify which Camera to use
        CameraSelector cameraSelector = new CameraSelector.Builder().requireLensFacing(CameraSelector.LENS_FACING_BACK).build();
        // Delay till start
        new CountDownTimer((long)(START_DELAY * 1000), 1000){
            @Override
            public void onTick(long millisUntilFinished) {}
            @Override
            public void onFinish() {
                cameraProvider.bindToLifecycle((LifecycleOwner) context, cameraSelector, imageAnalysis, preview);
            }
        }.start();
    }


    private Mat extractYChannel(@NonNull Image img) {
        ByteBuffer yBuffer = img.getPlanes()[0].getBuffer();
        byte[] yData = new byte[yBuffer.remaining()];
        yBuffer.get(yData);

        Mat yMat = new Mat(img.getHeight(), img.getWidth(), CvType.CV_8UC1);
        yMat.put(0, 0, yData);

        return yMat;
    }

    private Mat processImage(Mat currentImage){
        if (previousImage == null) {
            previousImage = currentImage;
            return null;
        }

        Mat mat = new Mat();
        currentImage = addGaussianBlur(currentImage, BLUR_KERNEL_SIZE);

        mat = thresholdPixels(currentImage, previousImage, PIXEL_THRESHOLD);
        mat = dilateNonZero(mat, DILATE_KERNEL_SIZE);
        mat = thresholdContourArea(mat, CONTOUR_THRESHOLD);

        previousImage = currentImage.clone();
        return mat;
    }

    private Mat alternativeProcessImage(Mat currentImage){
        if (previousImage == null) {
            previousImage = currentImage;
            return null;
        } else if (lastThresh == null) {
            lastThresh = thresholdPixels(currentImage, previousImage, PIXEL_THRESHOLD);
            return null;
        }

        while (frameCnt < 20){
            Mat thresh = thresholdPixels(currentImage, previousImage, PIXEL_THRESHOLD);
            Core.bitwise_or(thresh, lastThresh, lastThresh);
            frameCnt++;
            return currentOut;
        }
        return lastThresh;
    }


    private Mat thresholdPixels(Mat inputMat, Mat previousImage, float luminosityThreshold){
        Mat diffImage = new Mat();
        Core.absdiff(inputMat, previousImage, diffImage);

        Mat binaryMat = new Mat();
        Imgproc.threshold(diffImage, binaryMat, luminosityThreshold, 255, Imgproc.THRESH_BINARY);

        return binaryMat;
    }

    private Mat dilateNonZero(Mat inputMat, int kernelSize){
        Mat dilatedMat = new Mat();
        Mat kernel = Imgproc.getStructuringElement(Imgproc.MORPH_ELLIPSE, new org.opencv.core.Size(kernelSize, kernelSize));
        Imgproc.dilate(inputMat, dilatedMat, kernel);
        return dilatedMat;
    }

    private Mat addGaussianBlur(Mat inputMat, int kernelSize){
        Mat outputMat = new Mat();
        Imgproc.GaussianBlur(inputMat, outputMat, new org.opencv.core.Size(kernelSize, kernelSize), 0);
        return outputMat;
    }

    private Mat thresholdContourArea(Mat inputMat, float areaThreshold){
        List<MatOfPoint> contours = new ArrayList<>();
        Mat hierarchy = new Mat();
        Imgproc.findContours(inputMat, contours, hierarchy, Imgproc.RETR_EXTERNAL, Imgproc.CHAIN_APPROX_SIMPLE);

        Mat outputMat = new Mat(inputMat.size(), inputMat.type(), new Scalar(0));
        // Iterate over the contours and draw only the larger contours on the outputMat
        for (MatOfPoint contour : contours) {
            double contourArea = Imgproc.contourArea(contour);
            if (contourArea > areaThreshold) {
                Imgproc.drawContours(outputMat, Collections.singletonList(contour), 0, new Scalar(255), -1);
            }
        }
        // Apply the outputMat as a mask to the dilatedImage
        Mat maskedImage = new Mat();
        inputMat.copyTo(maskedImage, outputMat);
        return outputMat;
    }

    private int countNonZeroPixels(Mat inputImage) {
        return Core.countNonZero(inputImage);
    }

    private void debugMat(Mat mat, ImageView imageView) {
        if (imageView == null || mat == null)
            return;

        Bitmap bitmap = Bitmap.createBitmap(mat.cols(), mat.rows(), Bitmap.Config.ARGB_8888);
        Utils.matToBitmap(mat, bitmap);

        // Display the bitmap in an ImageView
        imageView.setImageBitmap(bitmap);
    }
}